'''
===================================
Author: Stavros Orfanoudakis 2023
===================================
'''

import numpy as np
import warnings
import math
from typing import Tuple, Union


class EV():
    '''
    This file contains the EV class, which is used to represent the EVs in the environment.
    The two-stage battery model was adapted from https://github.com/zach401/acnportal/blob/master/acnportal/acnsim/models/battery.py#L186


    Attributes:
        - id: unique identifier of the EV (uniquep per charging station)
        - location: the location of the EV (charging station id)
        - battery_capacity_at_arrival: the battery capacity of the EV at arrival time in kWh
        - time_of_arrival: the time of arrival of the EV in the charging station in simulation timesteps
        - time_of_departure: the earliest time of departure of the EV in the charging station in simulation timesteps (if use_probabilistic_time_of_departure is False, then time_of_departure is equal to time_of_departure)
        - desired_capacity: the desired capacity of the EV at departure time in kWh
        - use_probabilistic_time_of_departure: whether the EV will use a probabilistic time of departure or not
        - battery_capacity: the battery capacity of the EV in kWh
        - min_desired_capacity: the minimum desired capacity of the EV in kWh to maximize battery life
        - max_desired_capacity: the maximum desired capacity of the EV in kWh to maximize battery life
        - charge_efficiency: the efficiency of the EV when charging
        - discharge_efficiency: the efficiency of the EV when discharging        
        - timescale: the timescale of the simulation (useful for determining the charging speed)

    Status variables:
        - current_capacity: the current battery capacity of the EV in kWh
        - current_energy: the current power input of the EV in kW (positive for charging, negative for discharging)
        - charging_cycles: the number of charging/discharging cycles of the EV (useful for determining battery life parameters)
        - previous_power: the power input of the EV in the previous timestep in kW (positive for charging, negative for discharging)

    Methods:
        - step: updates the EV status according to the actions taken by the EV charger
        - _charge: charges the EV
        - _discharge: discharges the EV        

    '''

    def __init__(self,
                 id,
                 location,
                 battery_capacity_at_arrival,
                 time_of_arrival,
                 time_of_departure,                 
                 desired_capacity=None,  # kWh
                 battery_capacity=50,  # kWh
                 min_battery_capacity=10,  # kWh
                 max_ac_charge_power=22,  # kW
                 min_ac_charge_power=0,  # kW
                 max_dc_charge_power=50,  # kW
                 max_discharge_power=-22,  # kW
                 min_discharge_power=0,  # kW                 
                 ev_phases=3,                 
                 transition_soc=0.8,
                 charge_efficiency=1,
                 discharge_efficiency=1,                 
                 timescale=5,
                 ):

        self.id = id
        self.location = location
        self.timescale = timescale        

        # EV simulation characteristics
        self.time_of_arrival = time_of_arrival
        self.time_of_departure = time_of_departure        
        self.desired_capacity = battery_capacity if desired_capacity is None else desired_capacity
        self.battery_capacity_at_arrival = battery_capacity_at_arrival  # kWh

        # EV technical characteristics
        self.battery_capacity = battery_capacity  # kWh
        self.min_battery_capacity = min_battery_capacity  # kWh
        self.max_ac_charge_power = max_ac_charge_power  # kW
        self.min_ac_charge_power = min_ac_charge_power  # kW
        self.max_discharge_power = max_discharge_power  # kW
        self.min_discharge_power = min_discharge_power  # kW
        self.max_dc_charge_power = max_dc_charge_power  # kW
        self.transition_soc = transition_soc        
        self.ev_phases = ev_phases

        self.charge_efficiency = charge_efficiency
        self.discharge_efficiency = discharge_efficiency        

        # EV status
        self.current_capacity = battery_capacity_at_arrival  # kWh
        self.prev_capacity = self.current_capacity
        self.current_energy = 0  # kW
        self.actual_current = 0  # A
        self.charging_cycles = 0
        self.previous_power = 0
        self.required_energy = self.battery_capacity - self.battery_capacity_at_arrival
        self.total_energy_exchanged = 0

        # Baterry degradation        
        self.abs_total_energy_exchanged = 0
        self.historic_soc = []
        self.active_steps = []
        
        self.calendar_loss = 0
        self.cyclic_loss = 0

    def reset(self):
        '''
        The reset method is used to reset the EV's status to the initial state.
        '''
        self.current_capacity = self.battery_capacity_at_arrival
        self.current_energy = 0
        self.actual_current = 0
        self.charging_cycles = 0
        self.previous_power = 0
        self.required_energy = self.battery_capacity - self.battery_capacity_at_arrival
        self.c_lost = 0

    def step(self, amps, voltage, phases=1, type='AC') -> Tuple[float, float]:
        '''
        The step method is used to update the EV's status according to the actions taken by the EV charger.
        Inputs:
            - action: the power input in kW (positive for charging, negative for discharging)
        Outputs:
            - self.current_energy: the current power input of the EV in kW (positive for charging, negative for discharging)
            - self.actual_curent: the actual current input of the EV in A (positive for charging, negative for discharging)
        '''

        if type == 'DC':
            raise NotImplementedError

        if amps > 0 and amps < self.min_ac_charge_power*1000/(voltage*math.sqrt(phases)):
            amps = 0
        elif amps < 0 and amps > self.min_discharge_power*1000/(voltage*math.sqrt(phases)):
            amps = 0
            
        self.historic_soc.append(self.get_soc())

        
        if amps == 0:
            self.current_energy = 0
            self.actual_current = 0
            
            self.active_steps.append(0)
            return 0, 0

        # If the action is different than the previous action, then increase the charging cycles
        if self.previous_power == 0 or (self.previous_power/amps) < 0:
            self.charging_cycles += 1

        phases = min(phases, self.ev_phases)

        if amps > 0:
            self.actual_current = self._charge(amps, voltage, phases)
        elif amps < 0:
            self.actual_current = self._discharge(amps, voltage, phases)

        self.previous_power = self.current_energy

        self.total_energy_exchanged += self.current_energy #* self.timescale / 60
        self.abs_total_energy_exchanged += abs(self.current_energy) #* self.timescale / 60
        
        #round up to the nearest 0.01 the current capacity
        self.current_capacity = self.my_ceil(self.current_capacity, 2)
        
        self.active_steps.append(1 if self.actual_current != 0 else 0)
        return self.current_energy, self.actual_current

    def my_ceil(self, a, precision=2):
        return np.true_divide(np.ceil(a * 10**precision), 10**precision)

    def is_departing(self, timestep) -> Union[float, None]:
        '''
        The is_departing method is used to determine whether the EV is departing or not.
        Inputs:
            - timestep: the current timestep of the simulation
        Outputs:
            - Returns the user satisfaction of the EV in departing else None
        '''
        if timestep < self.time_of_departure:
            return None

        return self.get_user_satisfaction()

    def get_user_satisfaction(self) -> float:
        '''
        A function that returns the user satisfaction of the EV when departing.
        Outputs: 
            - Score: a value between 0 and 1
        '''

        if self.current_capacity < self.desired_capacity - 0.001:            
            return self.current_capacity / self.desired_capacity
        else:
            return 1

    def get_soc(self) -> float:
        '''
        A function that returns the state of charge of the EV.
        Outputs: 
            - SoC: the state of charge of the EV in [0,1]
        '''
        return (self.current_capacity/self.battery_capacity)

    def __str__(self):
        return f' {self.current_energy*60/self.timescale :5.1f} kWh |' + \
            f' {(self.current_capacity/self.battery_capacity)*100:5.1f} % |' + \
            f't_stay: {self.time_of_arrival}-' + \
            f'{self.time_of_departure} |' + \
            f' {self.max_ac_charge_power}/' + \
            f'{self.max_discharge_power} kWh|' + \
            f' {self.battery_capacity} kW |'

    def _charge(self, amps, voltage, phases=1) -> float:

        assert (amps > 0)
        # given_power = (amps * voltage / 1000) * \
        #     self.charge_efficiency * self.timescale / 60  # KW

        # if self.current_capacity + given_power > self.battery_capacity:
        #     self.current_energy = self.battery_capacity - self.current_capacity
        #     self.current_capacity = self.battery_capacity
        # else:
        #     self.current_energy = given_power
        #     self.current_capacity += given_power

        """ Method to "charge" the battery based on a two-stage linear
        battery model.
        
        Battery model based on a piecewise linear approximation of battery behavior. The battery will charge at the
        minimum of max_rate and the pilot until it reaches _transition_soc. After this, the maximum charging rate of the
        battery will decrease linearly to 0 at 100% state of charge.

        For more info on model: https://www.sciencedirect.com/science/article/pii/S0378775316317396

        All calculations are done in terms fo battery state of charge
        (SoC). Results are converted back to power units at the end.
        Code adapted from: https://github.com/zach401/acnportal/blob/master/acnportal/acnsim/models/battery.py#L186

        Args:
            pilot (float): Pilot signal passed to the battery. [A]
            voltage (float): AC voltage provided to the battery
                charger. [V]
            period (float): Length of the charging period. [minutes]

        Returns:
            float: average charging rate of the battery over this single
                period.

        """

        pilot = amps
        voltage = voltage * math.sqrt(phases)
        period = self.timescale
        # All calculations are done in terms of battery SoC, so we
        # convert pilot signal and max power into pilot and max rate of
        # change of SoC.
        pilot_dsoc = self.charge_efficiency * pilot * voltage / 1000 / \
            self.battery_capacity / (60 / period)
        max_dsoc = self.charge_efficiency * self.max_ac_charge_power / \
            self.battery_capacity / (60 / period)

        if pilot_dsoc > max_dsoc:
            pilot_dsoc = max_dsoc

        if self.transition_soc == 1:
            curr_soc = pilot_dsoc + self.get_soc()
            if curr_soc > 1:
                curr_soc = 1
        
        else:
            # The pilot SoC rate of change has a new transition SoC at
            # which decreasing of max charging rate occurs.
            pilot_transition_soc = self.transition_soc + (
                pilot_dsoc - max_dsoc
            ) / max_dsoc * (self.transition_soc - 1)

            # The charging equation depends on whether the current SoC of
            # the battery is above or below the new transition SoC.
            if self.get_soc() < pilot_transition_soc:
                # In the pre-rampdown region, the charging equation changes
                # depending on whether charging the battery over this
                # time period causes the battery to transition between
                # charging regions.
                if 1 <= (pilot_transition_soc - self.get_soc()) / pilot_dsoc:
                    curr_soc = pilot_dsoc + self.get_soc()
                else:
                    curr_soc = 1 + np.exp(
                        (pilot_dsoc + self.get_soc() - pilot_transition_soc)
                        / (pilot_transition_soc - 1)
                    ) * (pilot_transition_soc - 1)
            else:
                curr_soc = 1 + np.exp(pilot_dsoc / (pilot_transition_soc - 1)) * (
                    self.get_soc() - 1
                )

        dsoc = curr_soc - self.get_soc()
        self.prev_capacity = self.current_capacity
        self.current_capacity = curr_soc * self.battery_capacity

        # For charging power and charging rate (current), we use the
        # the average over this time period.
        self.current_energy = dsoc * self.battery_capacity  # / (period / 60)
        self.required_energy = self.required_energy - self.current_energy  # * period / 60
        return self.current_energy / (period / 60) * 1000 / voltage

    def _discharge(self, amps, voltage, phases) -> float:
        '''
        The _discharge method is used to discharge the EV's battery.
        Inputs:
            - power: the power input in kW (it is negative because of the discharge)
        '''
        assert (amps < 0)

        voltage = voltage * math.sqrt(phases)
        
        given_power = (amps * voltage / 1000)
        
        if abs(given_power/1000) > abs(self.max_discharge_power):
            given_power = self.max_discharge_power

        given_energy = given_power * self.discharge_efficiency * self.timescale / 60        
        if self.current_capacity + given_energy < self.min_battery_capacity:
            self.current_energy = -(self.current_capacity - self.min_battery_capacity)
            given_energy = self.current_energy
            self.prev_capacity = self.current_capacity
            self.current_capacity = self.min_battery_capacity
        else:
            self.current_energy = given_energy  # * 60 / self.timescale
            self.prev_capacity = self.current_capacity
            self.current_capacity += given_energy        
        
        self.required_energy = self.required_energy + self.current_energy

        return given_energy*60/self.timescale * 1000 / voltage

    def get_battery_degradation(self) -> Tuple[float, float]:
        '''
        A function that returns the capacity loss of the EV.

        Qacc := Accumulated battery cell throughput (Ah)
        Qsim := Battery cell throughput during simulation (Ah)        
        Tacc := Battery age (days)
        Tsim := Simulation time (days)
        theta := Battery temperature (K)

        Outputs: 
            - Capacity loss: the capacity loss
        '''

        # Degradation modelling parameters
        e0 = 7.543e6
        e1 = 23.75e6
        e2 = 6976

        z0 = 7.348e-3
        z1 = 3.667
        z2 = 7.6e-4
        z3 = 4.081e-3

        b_cap_ah = 2.05  # ah
        b_cap_kwh = 78  # kwh

        d_dist = 15000  # km
        b_age = 2*365  # days
        G = 0.186  # kwh/km

        # Age of the battery in days
        T_acc = b_age

        # Simulation time in days
        T_sim = (self.time_of_departure - self.time_of_arrival + 1)*self.timescale/ (60*24) # days

        theta = 298.15  # Kelvin
        k = 0.8263  # Volts

        v_min = 3.3324  # Volts
        self.historic_soc.append(self.get_soc()) # Add the final soc to the historic soc
        avg_soc = np.mean(self.historic_soc) 
        v_avg = v_min + k * avg_soc

        # alpha(v_avg)
        alpha = (e0 * v_avg - e1) * math.exp(-e2 / theta)
        d_cal = alpha * 0.75 * T_sim / (T_acc)**0.25

        # beta(v_avg, soc_avg)
        # print(f'avg_soc: {avg_soc}')        
        self.active_steps.append(1)
        
        # get historic soc that self.active_steps == 1
        filtered_historic_soc = [soc for i, soc in enumerate(self.historic_soc) if self.active_steps[i] == 1]
        # print(f'filtered soc {filtered_historic_soc}')
        avg_filtered_soc = np.mean(filtered_historic_soc) 
        
        delta_DoD = 2 * abs(avg_filtered_soc.repeat(len( filtered_historic_soc)) - filtered_historic_soc).mean()
        # print(f'delta_DoD: {delta_DoD}')
        v_half_soc = v_min + k * 0.5
        beta = z0 * (v_half_soc - z1)**2 + z2 + z3 * delta_DoD        

        Q_sim = (self.abs_total_energy_exchanged / b_cap_kwh) * b_cap_ah

        # accumulated throughput
        Q_acc = 2 * (b_age * (d_dist / 365) * G * b_cap_ah) / b_cap_kwh
        # print(f'Q_acc: {Q_acc}')

        d_cyc = beta * 0.5 * Q_sim / (Q_acc)**0.5

        self.calendar_loss = d_cal
        self.cyclic_loss = d_cyc
        
        return d_cal, d_cyc



# This file implements functions that simulate the grid of the simulations using PandaPower
#### NOT used yet, it is in development phase!!!!!!!!!!!!!!
from pandapower.plotting.plotly import pf_res_plotly
import pandapower as pp
import numpy as np

from pandapower.plotting.plotly.mapbox_plot import set_mapbox_token
set_mapbox_token(
    'pk.eyJ1Ijoic29yZmFub3VkYWtpcyIsImEiOiJjbGs2ejQycGcwM3dwM21ydDd2dGc1dG1oIn0.dJ9E-1u53SZfDH0xgVaBkw')


class Grid:

    def __init__(self, charging_stations, seed=9, case="default"):

        self.charging_stations = charging_stations
        # Load the grid case
        if case == "default":
            self.net = pp.networks.create_cigre_network_mv(with_der='pv_wind')
        else:
            raise NotImplementedError

        self.seed = seed
        np.random.seed(seed)

        # Initialize the Charging Stations
        for i in range(len(self.net.bus)):
            # Add a Storage Unit to each bus of the grid except the external grid bus
            if i not in self.net.ext_grid.bus.values:
                pp.create_storage(self.net, bus=i, p_mw=0, max_e_mwh=1000,
                                  q_mvar=0,  type="ChargingStation")
        # Initialize the charging stations conneted bus
        # get charging_stations random integers excluding the integers in self.net.ext_grid.bus.values
        self.charging_stations_buses = np.random.choice(
            [i for i in range(len(self.net.bus)) if i not in self.net.ext_grid.bus.values], charging_stations, replace=True)

    def _get_grid_actions(self,actions):
        # Transform ations to the format (number_of_ports,1) -> (bus, p_mw)

        return actions
    
    def step(self, actions):

        actions = _get_grid_actions(actions)

        # reset CS actions
        for i in range(len(self.net.storage)):
            self.net.storage.p_mw[i] = 0

        # actions are in the format (bus, p_mw)
        assert (len(actions) == self.charging_stations)
        for action in actions:
            self.net.storage.p_mw[action[0]] += action[1]

        std = 0.1
        # TODO: replace with REAL load profiles
        self.net.load.p_mw = np.random.normal(
            self.net.load.p_mw, std, (len(self.net.load)))
        self.net.load.q_mvar = np.random.normal(
            self.net.load.q_mvar, std, (len(self.net.load)))

        # TODO: replace with REAL static generation (pv, wind) profiles
        self.net.sgen.p_mw = np.random.normal(
            self.net.sgen.p_mw, std, (len(self.net.sgen)))
        # self.net.sgen.q_mvar = np.random.normal(self.net.sgen.q_mvar, std, (len(self.net.sgen)))

        # TODO: replace with REAL generation profiles
        self.net.gen.p_mw = np.random.normal(
            self.net.gen.p_mw, std, (len(self.net.gen)))
        self.net.gen.vm_pu = np.random.normal(
            self.net.gen.vm_pu, std, (len(self.net.gen)))

        # TODO: replace with REAL ext_grid profile
        self.net.ext_grid.vm_pu = np.random.normal(
            self.net.ext_grid.vm_pu, std, (len(self.net.ext_grid)))
        self.net.ext_grid.va_degree = np.random.normal(
            self.net.ext_grid.va_degree, std, (len(self.net.ext_grid)))

        # solve the power flow
        pp.runpp(self.net, numba=False)
    
    def get_grid_state(self):
        '''
        TODO: return the grid state
        '''
        pass

    def get_charging_stations_buses(self):
        return self.charging_stations_buses

    # TODO: this functino returns all the related transformers to the buses of the charging stations,
    #  it will help in the future to get the transformer loading objectives
    def get_bus_transformers(self):
        # return (n_charging_stations) with calues from the set {n_transformers}
        return None

    def visualize_pf(self):

        # visualize the grid using plotly
        # pp.plotting.plotly.simple_plotly(net)
        pf_res_plotly(self.net)
        # simple plot of net
        # pp.plotting.simple_plot(self.net, plot_loads=True, plot_sgens=True)

    def get_overloaded_lines(self):
        return self.net.res_line[self.net.res_line.loading_percent > 100]

    def get_overloaded_trafos(self, verbose=True):
        if verbose:
            print(
                f'Traffo loading percent:\n {self.net.res_trafo.loading_percent}')
        return self.net.res_trafo[self.net.res_trafo.loading_percent > 100]


if __name__ == "__main__":

    grid = Grid(4)

    timeslot_duration = 5  # minutes
    simulation_duration = 3  # * timeslot_duration minutes

    for simu_step in range(simulation_duration):
        print(f"Simulation step {simu_step}")
        actions = [(1, 0.5), (2, 0.1), (2, -0.2), (3, 0.1)]
        grid.step(actions)
        # grid.visualize_pf()

        # print information about the overaloading and the lines and the transformer from here
        # print(grid.get_overloaded_lines())
        grid.get_overloaded_trafos(verbose=False)

    # Get informatino about the overaloading and the lines and the transformer from here
    # print(net.res_line)
    # print(net.res_trafo)
    # print(net.res_trafo.loading_percent.round(3).astype(str))
    # # visualize the grid using plotly
    # pp.plotting.plotly.simple_plotly(net)

    # net = grid.net
    # print(
    #     f'Overloaded lines: {net.res_line[net.res_line.loading_percent > 100]}')
    # pp.plotting.to_html(net, "./test.html")

    # pf_res_plotly(net)

    # # simple plot of net
    # pp.plotting.simple_plot(net, plot_loads=True, plot_sgens=True)



from gymnasium.envs.registration import register

register(
    id='EV2Gym-v1',
    entry_point='ev2gym.models.ev2gym_env:EV2Gym',
    kwargs={'config_file': 'ev2gym/example_config_files/V2GProfitMax.yaml'}
)


# This file contains the state function for the RL agent
import math
import numpy as np


def PublicPST(env, *args):
    # This state function is the public power setpoints
    # The state is the public power setpoints
    # The state is a vector

    state = [
        (env.current_step/env.simulation_length),
        # env.sim_date.weekday() / 7,
        # turn hour and minutes in sin and cos
        # math.sin(env.sim_date.hour/24*2*math.pi),
        # math.cos(env.sim_date.hour/24*2*math.pi),
    ]

    # the final state of each simulation
    # if env.current_step < env.simulation_length:        
    #     setpoint = min(env.power_setpoints[env.current_step], env.charge_power_potential[env.current_step])        
    # else:
    #     setpoint = 0       
    if env.current_step < env.simulation_length:  
        # setpoint = env.power_setpoints[env.current_step:env.current_step+10]
        setpoint = env.power_setpoints[env.current_step]
    else:
        setpoint = np.zeros((1))
        
    # if len(setpoint) < 10:
    #     setpoint = np.append(setpoint, np.zeros(10-len(setpoint)))
    
    state.append(setpoint)
    state.append(env.current_power_usage[env.current_step-1])

    # For every transformer
    for tr in env.transformers:
        # For every charging station connected to the transformer
        for cs in env.charging_stations:
            if cs.connected_transformer == tr.id:
                # For every EV connected to the charging station
                for EV in cs.evs_connected:
                    # If there is an EV connected
                    if EV is not None:
                        state.append([
                            1 if EV.get_soc() == 1 else 0.5,  # we know if the EV is full
                            EV.total_energy_exchanged,
                            # EV.max_ac_charge_power*1000 /
                            # (cs.voltage*math.sqrt(cs.phases))/100,
                            # EV.min_ac_charge_power*1000 /
                            # (cs.voltage*math.sqrt(cs.phases))/100,
                            (env.current_step-EV.time_of_arrival)
                            ])

                    # else if there is no EV connected put zeros
                    else:
                        state.append(np.zeros(3))

    state = np.array(np.hstack(state))

    np.set_printoptions(suppress=True)

    return state

def V2G_profit_max(env, *args):

    state = [
        (env.current_step),        
    ]

    state.append(env.current_power_usage[env.current_step-1])

    charge_prices = abs(env.charge_prices[0, env.current_step:
        env.current_step+20])
    
    if len(charge_prices) < 20:
        charge_prices = np.append(charge_prices, np.zeros(20-len(charge_prices)))
    
    state.append(charge_prices)
    
    # For every transformer
    for tr in env.transformers:

        # For every charging station connected to the transformer
        for cs in env.charging_stations:
            if cs.connected_transformer == tr.id:

                # For every EV connected to the charging station
                for EV in cs.evs_connected:
                    # If there is an EV connected
                    if EV is not None:
                        state.append([
                            EV.get_soc(),
                            EV.time_of_departure - env.current_step,
                            ])

                    # else if there is no EV connected put zeros
                    else:
                        state.append(np.zeros(2))

    state = np.array(np.hstack(state))

    return state

def V2G_profit_max_loads(env, *args):
    '''
    This is the state function for the V2GProfitMax scenario with loads
    '''
    
    state = [
        (env.current_step),        
    ]

    state.append(env.current_power_usage[env.current_step-1])

    charge_prices = abs(env.charge_prices[0, env.current_step:
        env.current_step+20])
    
    if len(charge_prices) < 20:
        charge_prices = np.append(charge_prices, np.zeros(20-len(charge_prices)))
    
    state.append(charge_prices)
    
    # For every transformer
    for tr in env.transformers:
        loads, pv = tr.get_load_pv_forecast(step = env.current_step,
                                            horizon = 20)
        power_limits = tr.get_power_limits(step = env.current_step,
                                           horizon = 20)
        state.append(loads-pv)
        state.append(power_limits)
        
        # For every charging station connected to the transformer
        for cs in env.charging_stations:
            if cs.connected_transformer == tr.id:

                # For every EV connected to the charging station
                for EV in cs.evs_connected:
                    # If there is an EV connected
                    if EV is not None:
                        state.append([
                            EV.get_soc(),
                            EV.time_of_departure - env.current_step,
                            ])

                    # else if there is no EV connected put zeros
                    else:
                        state.append(np.zeros(2))

    state = np.array(np.hstack(state))

    return state
    


def BusinessPSTwithMoreKnowledge(env, *args):
    '''
    This state function is used for the business case scenario that requires more knowledge such as SoC and time of departure for each EV present.
    '''

    state = [
        (env.current_step) / env.simulation_length,
        #env.sim_date.weekday() / 5,
        # turn hour and minutes in sin and cos
        #math.sin(env.sim_date.hour/12*2*math.pi),
        #math.cos(env.sim_date.hour/12*2*math.pi),
    ]

    # the final state of each simulation
    if env.current_step < env.simulation_length:
        state.append(env.power_setpoints[env.current_step]) #/100
        state.append(env.charge_power_potential[env.current_step]) #/100
    else:
        state.append(env.power_setpoints[env.current_step-1]) #/100
        state.append(env.charge_power_potential[env.current_step-1]) #/100   

    for tr in env.transformers:
        state.append(tr.max_current/100)
        for cs in env.charging_stations:
            if cs.connected_transformer == tr.id:
                for EV in cs.evs_connected:
                    if EV is not None:
                        state.append([#EV.total_energy_exchanged / EV.battery_capacity, #how much soc we charge
                                      #EV.max_ac_charge_power*1000 /            same EVs, no need right now
                                      #(cs.voltage*math.sqrt(cs.phases)),
                                      #EV.min_ac_charge_power*1000 /
                                      #(cs.voltage*math.sqrt(cs.phases)),
                                      EV.time_of_arrival / env.simulation_length,  # time of arrival
                                      EV.etime_of_departure / env.simulation_length,  # time of departure
                                      EV.get_soc(),  # soc
                                      #(EV.etime_of_departure - env.current_step) \
                                      #  / env.simulation_length, #remaining time
                                      #(env.current_step-EV.time_of_arrival) \
                                      #  / env.simulation_length,  # time stayed
                                      #(EV.etime_of_departure - \
                                      # EV.time_of_arrival) / env.simulation_length, # total staying time
                                      #(((EV.battery_capacity - EV.battery_capacity_at_arrival) /
                                      #  (EV.etime_of_departure - EV.time_of_arrival)) / EV.max_ac_charge_power),  # average charging speed
                                      #(((EV.battery_capacity - EV.battery_capacity_at_arrival) / EV.battery_capacity)) \
                                      #  / ((EV.etime_of_departure - env.current_step + 1) / env.simulation_length),   #charging priority
                                      #EV.required_power / EV.battery_capacity,  # required energy
                                      ])
                    else:
                        state.append(np.zeros(3))

    state = np.array(np.hstack(state))

    np.set_printoptions(suppress=True)

    return state


'''
===================================
Author: Stavros Orfanoudakis 2023
===================================
'''
import numpy as np


class Transformer():
    """
    Transformer class for the ev_city environment


    """

    def __init__(self,
                 id,  # unique identifier of the transformer
                 env,
                 max_power=100,  # The maximum power of the transformer in kW
                 cs_ids=[],  # the charging stations connected to the transformer
                 inflexible_load=np.zeros(96),
                 solar_power=np.zeros(96),
                 simulation_length=96
                 ):
        """
        Initialize the transformer

        :param id: unique identifier of the transformer
        :type id: int
        :param min_current: The minimum capacity of the transformer in A, defaults to 0
        :type min_current: int, optional
        :param cs_ids: the charging stations connected to the transformer, defaults to []
        :type cs_ids: list, optional
        :param timescale: the timescale of the simulation, defaults to 5
        :type timescale: int, optional

        """

        self.id = id
        max_current = max_power * 1000 / 400
        self.max_current = np.ones(simulation_length)*max_current
        self.min_current = np.ones(simulation_length) * -max_current
        self.max_power = np.ones(simulation_length)*max_power
        self.min_power = np.ones(simulation_length) * -max_power

        self.inflexible_load = inflexible_load
        self.solar_power = solar_power

        self.cs_ids = cs_ids
        self.simulation_length = simulation_length

        self.current_amps = 0
        self.current_power = 0

        self.current_step = 0

        self.inflexible_load_forecast = np.zeros(env.simulation_length)
        if env.config['inflexible_loads']['include']:
            self.normalize_inflexible_loads(env)
            self.generate_inflexible_loads_forecast(env)

        self.pv_generation_forecast = np.zeros(96)
        if env.config['solar_power']['include']:
            self.normalize_pv_generation(env)
            self.generate_pv_generation_forecast(env)

        self.steps_ahead = env.config['demand_response']['notification_of_event_minutes'] // env.timescale

        if env.config['demand_response']['include']:
            self.dr_events = self.generate_demand_response_events(env)
            # clip the forecast to the min and max power
            self.inflexible_load_forecast = np.clip(self.inflexible_load_forecast,
                                                    self.min_power,
                                                    self.max_power)
        else:
            self.dr_events = []

    def generate_demand_response_events(self, env) -> None:
        '''
        This function is used to generate demand response events using the configuration file
        and by updating the transformer loading
        '''
        events_per_day = env.config['demand_response']['events_per_day']

        event_length_minutes_min = env.config['demand_response']['event_length_minutes_min']
        event_length_minutes_max = env.config['demand_response']['event_length_minutes_max']

        event_start_hour_mean = env.config['demand_response']['event_start_hour_mean']
        event_start_hour_std = env.config['demand_response']['event_start_hour_std']
        event_capacity_percentage_mean = env.config['demand_response']['event_capacity_percentage_mean']
        event_capacity_percentage_std = env.config['demand_response']['event_capacity_percentage_std']

        events = []
        for _ in range(events_per_day):

            event_length_minutes = env.tr_rng.integers(
                event_length_minutes_min, event_length_minutes_max+1)

            event_start_hour = env.tr_rng.normal(
                event_start_hour_mean*60, event_start_hour_std*60)

            event_start_hour = np.clip(event_start_hour, 0, 23*60)
            event_start_step = event_start_hour // env.timescale

            sim_start_step = (env.sim_date.hour * 60 +
                              env.sim_date.minute) // env.timescale
            event_start_step = int(event_start_step - sim_start_step)

            event_end_step = int(event_start_step +
                                 event_length_minutes // env.timescale)

            capacity_percentage = env.tr_rng.normal(
                event_capacity_percentage_mean, event_capacity_percentage_std)
            capacity_percentage = np.clip(capacity_percentage, 0, 100)

            self.max_power[event_start_step:event_end_step] = self.max_power[event_start_step:event_end_step] - \
                (self.max_power[event_start_step:event_end_step] *
                    capacity_percentage/100)

            self.max_current[event_start_step:event_end_step] = self.max_current[event_start_step:event_end_step] - \
                (self.max_current[event_start_step:event_end_step] *
                    capacity_percentage/100)

            if any(self.inflexible_load[event_start_step:event_end_step] >
                   self.max_power[event_start_step:event_end_step]):

                self.max_power[event_start_step:event_end_step] = self.inflexible_load[event_start_step:
                                                                                       event_end_step].max()

                capacity_percentage = 100*(1 - max(self.inflexible_load[event_start_step:event_end_step]) /
                                           max(self.max_power))

            event = {'event_start_step': event_start_step,
                     'event_end_step': event_end_step,
                     'capacity_percentage': capacity_percentage}
            events.append(event)

        return events

    def get_power_limits(self, step, horizon) -> np.array:
        '''
        Get the power limits of the transformer for the next horizon steps.

        If an event is happening, the power limits are updated accordingly
        The control algorithm is aware of the event x steps before it happens

        '''
        known_max_power = max(self.max_power) * np.ones(horizon)
        power_limit = max(self.max_power)

        for event in self.dr_events:
            if step + self.steps_ahead >= event['event_start_step'] and \
                    event['event_end_step'] >= step:

                if step > event['event_start_step']:
                    known_max_power[:event['event_end_step']-step] = power_limit - \
                        power_limit * event['capacity_percentage'] / 100
                else:
                    known_max_power[abs(event['event_start_step'] - step):
                                    abs(event['event_end_step']-step)] = power_limit - \
                        power_limit * event['capacity_percentage'] / 100
                continue

        # append the last value if the horizon is larger than the max_power
        if len(known_max_power) < horizon:
            known_max_power = np.append(known_max_power, np.ones(horizon-len(known_max_power))
                                        * power_limit)

        return known_max_power

    def get_load_pv_forecast(self, step, horizon) -> np.array:

        load_forecast = self.inflexible_load_forecast[step:step+horizon]
        pv_forecast = self.pv_generation_forecast[step:step+horizon]
        
        if step < len(self.inflexible_load_forecast):                        
            load_forecast[0] = self.inflexible_load[step]
            pv_forecast[0] = self.solar_power[step]

        if len(load_forecast) < horizon:
            load_forecast = np.append(load_forecast, np.zeros(horizon-len(load_forecast))
                                      * self.inflexible_load_forecast[-1])
            pv_forecast = np.append(pv_forecast, np.zeros(horizon-len(pv_forecast))
                                    * self.pv_generation_forecast[-1])

        return load_forecast, pv_forecast

    def normalize_pv_generation(self, env) -> None:
        '''
        Normalize the solar_power using the configuration file and teh max_power of the transformer
        '''
        if env.config['solar_power']['include']:
            mult = env.config['solar_power']['solar_power_capacity_multiplier_mean']
            mult = env.tr_rng.normal(mult, 0.1)
            self.solar_power = -self.solar_power * \
                mult * max(self.max_power)

    def generate_pv_generation_forecast(self, env) -> None:
        '''
        This function is used to generate pv generation forecast using the configuration file
        '''
        forecast_uncertainty_mean = env.config['solar_power']['forecast_mean'] / 100 * \
            self.solar_power

        forecast_uncertainty_std = env.config['solar_power']['forecast_std'] / 100 * \
            self.solar_power

        self.pv_generation_forecast = np.random.normal(
            forecast_uncertainty_mean,
            abs(forecast_uncertainty_std),
            len(self.solar_power))

    def normalize_inflexible_loads(self, env) -> None:
        '''
        Check that infelxible_loads are lower than the max_power, if not, set them to the max_power
        '''

        if env.config['inflexible_loads']['include']:
            mult = env.config['inflexible_loads']['inflexible_loads_capacity_multiplier_mean']
            mult = env.tr_rng.normal(mult, 0.1)

            # scale up the data to match the max_power of the transformers
            self.inflexible_load = self.inflexible_load * \
                mult * (max(self.max_power) /
                        self.inflexible_load.max()+0.0000001)
            # for each step
            for j in range(env.simulation_length):
                if self.inflexible_load[j] > self.max_power[j]:
                    self.inflexible_load[j] = self.max_power[j]

                elif self.inflexible_load[j] < self.min_power[j]:
                    self.inflexible_load[j] = self.min_power[j]

        self.generate_inflexible_loads_forecast(env)

    def generate_inflexible_loads_forecast(self, env) -> None:
        '''
        This function is used to generate inflexible loads forecast using the configuration file
        '''
        forecast_uncertainty_mean = env.config['inflexible_loads']['forecast_mean'] / 100 * \
            self.inflexible_load

        forecast_uncertainty_std = env.config['inflexible_loads']['forecast_std'] / 100 * \
            self.inflexible_load

        self.inflexible_load_forecast = np.random.normal(
            forecast_uncertainty_mean,
            forecast_uncertainty_std,
            len(self.inflexible_load))

        # clip the forecast to the min and max power
        self.inflexible_load_forecast = np.clip(self.inflexible_load_forecast,
                                                self.min_power,
                                                self.max_power)

    def reset(self, step) -> None:
        '''
        Reset the current power of the transformer
        '''
        self.current_step = step

        self.current_power = self.inflexible_load[step] + \
            self.solar_power[step]

        self.current_amps = (self.current_power * 1000) / 400

    def step(self, amps, power) -> None:
        '''
        Update current power of the transformer
        '''
        self.current_amps += amps
        self.current_power += power

    def is_overloaded(self) -> bool:
        '''
        Check if the transformer is overloaded

        Returns:
            - True if the transformer is overloaded
            - False if the transformer is not overloaded
        '''
        e = 0.0001

        if self.current_power > self.max_power[self.current_step] + e \
                or self.current_power < self.min_power[self.current_step] - e:
            return True

        return False

    def get_how_overloaded(self) -> float:
        '''
        Check how overloaded the transformer is

        Returns:
            - a amps value if the transformer is overloaded
        '''
        if self.is_overloaded():
            return np.abs(self.current_power - self.max_power[self.current_step])
        else:
            return 0

    def __str__(self) -> str:

        return f'  - Transformer {self.id}:  {self.min_power[self.current_step]:.1f} / ' +\
            f'{self.current_power:5.1f} (L:{self.inflexible_load[self.current_step]:5.1f},' +\
            f' PV: {self.solar_power[self.current_step]:5.1f},' + \
            f' EVs: {(self.current_power-self.inflexible_load[self.current_step] - self.solar_power[self.current_step]):5.1f},' + \
            f'{self.max_power[self.current_step]:5.1f} kW' +\
            f'\tCSs: {self.cs_ids}'



# This is a setup file for the package

from setuptools import find_packages
from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='ev2gym',
    version='1.0.0',
    description='A realistic V2G simulator environment',
    author='Stavros Orfanoudakis',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='gym, Reinforcement Learning, V2X, V2G, G2V, EVs, ev2gym, Electric Vehicles, Electric Vehicle Simulator',
    # package_dir = {"": "ev2gym"},
    packages=find_packages(),
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={
        '': ['*.csv','*.json','*.yaml','*.npy','*.png']
    },
    include_package_data=True,
    install_requires=[
        'gymnasium',
        'pyyaml',
        'matplotlib',
        'pandas',
        'networkx',
        'gurobipy',
    ]
)

"""
rm -rf build dist *.egg-info
py -m build
python3 -m twine upload --repository pypi dist/*
username: __token__
use pypy-api-token
"""




'''
This file is used to plot the comparatigve results of the different algorithms.
'''

import pickle
import math
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import datetime
import os
from ev2gym.models.ev2gym_env import EV2Gym


marker_list = ['.', 'x', 'o', 'v', 's', 'p',
               'P', '*', 'h', 'H', '+', 'X', 'D', 'd', '|', '_']

# color_list = ['#00429d', '#5681b9', '#93c4d2', '#ffa59e', '#dd4c65', '#93003a']

color_list = ['#00429d', '#5681b9', '#93c4d2', '#ffa59e',
              '#dd4c65', '#93003a', 'b', 'g', 'r', 'c', 'm', 'y', 'k']

algorithm_names = [
    'Charge As Fast As Possible',
    # 'Charge As Late As Possible',
    # 'Round Robin',
    'OCCF V2G',
    'OCCF G2V',
    'eMPC V2G',
    'eMPC G2V',
]


def plot_total_power(results_path, save_path=None, algorithm_names=None):

    # Load the env pickle files
    with open(results_path, 'rb') as f:
        replay = pickle.load(f)

    plt.figure(figsize=(10, 7))
    plt.rc('font', family='serif')
    light_blue = np.array([0.529, 0.808, 0.922, 1])
    gold = np.array([1, 0.843, 0, 1])

    for index, key in enumerate(replay.keys()):        
        env = replay[key]

        date_range = pd.date_range(start=env.sim_starting_date,
                                   end=env.sim_starting_date +
                                   (env.simulation_length - 1) *
                                   datetime.timedelta(
                                       minutes=env.timescale),
                                   freq=f'{env.timescale}min')
        date_range_print = pd.date_range(start=env.sim_starting_date,
                                         end=env.sim_date,
                                         periods=10)

        counter = 1
        dim_x = int(np.ceil(np.sqrt(env.number_of_transformers)))
        dim_y = int(np.ceil(env.number_of_transformers/dim_x))
        for tr in env.transformers:

            plt.subplot(dim_x, dim_y, counter)
            df = pd.DataFrame([], index=date_range)

            if env.config['inflexible_loads']['include']:
                df['inflexible'] = env.tr_inflexible_loads[tr.id, :]
            if env.config['solar_power']['include']:
                df['solar'] = env.tr_solar_power[tr.id, :]

            for cs in tr.cs_ids:
                df[cs] = env.cs_power[cs, :]

            if index == 0:
                # plot the inflexible loads as a fill between
                if env.config['inflexible_loads']['include']:
                    plt.fill_between(df.index,
                                     np.array([0]*len(df.index)),
                                     df['inflexible'],
                                     step='post',
                                     alpha=0.7,
                                     color=light_blue,
                                     linestyle='--',
                                     linewidth=2,
                                     label='Inflexible Loads')

                # plot the solar power as a fill between the inflexible loads and the solar power
                if env.config['solar_power']['include']:
                    plt.fill_between(df.index,
                                     df['inflexible'],
                                     df['solar'] + df['inflexible'],
                                     step='post',
                                     alpha=0.7,
                                     color=gold,
                                     linestyle='--',
                                     linewidth=2,
                                     label='Solar Power')

                if env.config['demand_response']['include']:
                    plt.fill_between(df.index,
                                     np.array([tr.max_power.max()]
                                              * len(df.index)),
                                     tr.max_power,
                                     step='post',
                                     alpha=0.7,
                                     color='r',
                                     hatch='xx',
                                     linestyle='--',
                                     linewidth=2,
                                     label='Demand Response Event')

                plt.step(df.index,
                         #  tr.max_power
                         [tr.max_power.max()] * len(df.index),
                         where='post',
                         color='r',
                         linestyle='--',
                         linewidth=2,
                         label='Transformer Max Power')
                plt.plot([env.sim_starting_date, env.sim_date],
                         [0, 0], 'black')

            df['total'] = df.sum(axis=1)

            # plot total and use different color and linestyle for each algorithm
            plt.step(df.index, df['total'],
                     color=color_list[index],
                     where='post',
                     linestyle='-',
                     linewidth=1,
                     marker=marker_list[index],
                     label=algorithm_names[index])

            counter += 1

    plt.title(f'Transformer {tr.id+1}', fontsize=28)
    plt.xlabel(f'Time', fontsize=28)
    plt.ylabel(f'Power (kW)', fontsize=28)
    plt.xlim([env.sim_starting_date, env.sim_date])
    plt.xticks(ticks=date_range_print,
               labels=[
                   f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print],
               rotation=45,
               fontsize=28)
    plt.yticks(fontsize=28)
    # put legend under the plot
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.15),
               fancybox=True, shadow=True, ncol=3, fontsize=24)

    plt.grid(True, which='minor', axis='both')
    plt.tight_layout()

    fig_name = f'{save_path}/Transformer_Aggregated_Power.png'
    plt.savefig(fig_name, format='png',
                dpi=60, bbox_inches='tight')


def plot_total_power_V2G(results_path, save_path=None, algorithm_names=None):

    # Load the env pickle files
    with open(results_path, 'rb') as f:
        replay = pickle.load(f)

    plt.close('all')
    fig, ax = plt.subplots()
    plt.grid(True, which='major', axis='both')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    
    ax.spines['left'].set_linewidth(2)
    # ax.spines['bottom'].set_linewidth(2)           
    
    plt.rc('font', family='serif')
   
    light_blue = np.array([0.529, 0.808, 0.922, 1])
    gold = np.array([1, 0.843, 0, 1])
    
    color_list_map = plt.cm.get_cmap('Set1', len(replay.keys()))
    color_list = color_list_map(np.linspace(0, 1, len(replay.keys())))

    for index, key in enumerate(replay.keys()):        
        env = replay[key]

        date_range = pd.date_range(start=env.sim_starting_date,
                                   end=env.sim_starting_date +
                                   (env.simulation_length - 1) *
                                   datetime.timedelta(
                                       minutes=env.timescale),
                                   freq=f'{env.timescale}min')
        date_range_print = pd.date_range(start=env.sim_starting_date,
                                         end=env.sim_date,
                                         periods=7)

        counter = 1
        dim_x = int(np.ceil(np.sqrt(env.number_of_transformers)))
        dim_y = int(np.ceil(env.number_of_transformers/dim_x))
        for tr in env.transformers:

            plt.subplot(dim_x, dim_y, counter)
            df = pd.DataFrame([], index=date_range)

            if env.config['inflexible_loads']['include']:
                df['inflexible'] = env.tr_inflexible_loads[tr.id, :]
            if env.config['solar_power']['include']:
                df['solar'] = env.tr_solar_power[tr.id, :]

            for cs in tr.cs_ids:
                df[cs] = env.cs_power[cs, :]

            if index == 0:
                # plot the inflexible loads as a fill between
                if env.config['inflexible_loads']['include']:
                    plt.fill_between(df.index,
                                     np.array([0]*len(df.index)),
                                     df['inflexible'],
                                     step='post',
                                     alpha=0.3,
                                     color=light_blue,
                                     linestyle='--',
                                     linewidth=2,
                                     label='Inflexible Loads')

                # plot the solar power as a fill between the inflexible loads and the solar power
                if env.config['solar_power']['include']:
                    plt.fill_between(df.index,
                                     df['inflexible'],
                                     df['solar'] + df['inflexible'],
                                     step='post',
                                     alpha=0.8,
                                     color=gold,
                                     linestyle='--',
                                     linewidth=2,
                                     label='Solar Power')

                if env.config['demand_response']['include']:
                    plt.fill_between(df.index,
                                     np.array([tr.max_power.max()]
                                              * len(df.index)),
                                     tr.max_power,
                                     step='post',
                                     alpha=0.7,
                                     color='r',
                                     hatch='xx',
                                     linestyle='--',
                                     linewidth=2,
                                     label='Demand Response Event')

                plt.step(df.index,
                         #  tr.max_power
                         [-tr.max_power.max()] * len(df.index),
                         where='post',
                         color='r',
                         linestyle='--',
                         linewidth=2,
                         alpha=0.7,
                        #  label='Transf. Limit'
                         )
                
                plt.step(df.index,
                         #  tr.max_power
                         [tr.max_power.max()] * len(df.index),
                         where='post',
                         color='r',
                         linestyle='--',
                         linewidth=2,
                         alpha=0.7,
                         label='Transf. Limit')
                plt.plot([env.sim_starting_date, env.sim_date],
                         [0, 0], 'black')

            df['total'] = df.sum(axis=1)

            # plot total and use different color and linestyle for each algorithm
            plt.step(df.index, df['total'],
                     color=color_list[index],
                     where='post',
                     linestyle='-',
                     linewidth=1,
                     marker=marker_list[index],
                     label=algorithm_names[index])

            counter += 1
    
    plt.ylabel(f'Power (kW)', fontsize=28)
    plt.xlim([env.sim_starting_date, env.sim_date])
    plt.xticks(ticks=date_range_print,
               labels=[
                   f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print],
            #    rotation=45,
               fontsize=28)
    plt.yticks(fontsize=28)
    # put legend under the plot
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.1),
               fancybox=True, shadow=True, ncol=3, fontsize=24)    

    fig_name = f'{save_path}/Transformer_Aggregated_Power_Prices.png'
    
    plt.savefig(fig_name, format='png',
                dpi=60, bbox_inches='tight')

    plt.show()

def plot_comparable_EV_SoC(results_path, save_path=None, algorithm_names=None):
    '''
    This function is used to plot the SoC of the EVs in the same plot
    '''

    with open(results_path, 'rb') as f:
        replay = pickle.load(f)

    plt.close('all')
    plt.figure(figsize=(10, 7))
    plt.rc('font', family='serif')

    for index, key in enumerate(replay.keys()):        
        env = replay[key]

        date_range = pd.date_range(start=env.sim_starting_date,
                                   end=env.sim_starting_date +
                                   (env.simulation_length - 1) *
                                   datetime.timedelta(
                                       minutes=env.timescale),
                                   freq=f'{env.timescale}min')
        date_range_print = pd.date_range(start=env.sim_starting_date,
                                         end=env.sim_date,
                                         periods=10)

        counter = 1
        dim_x = int(np.ceil(np.sqrt(env.cs)))
        dim_y = int(np.ceil(env.cs/dim_x))
        for cs in env.charging_stations:            
            
            plt.subplot(dim_x, dim_y, counter)
            df = pd.DataFrame([], index=date_range)

            for port in range(cs.n_ports):
                df[port] = env.port_energy_level[port, cs.id, :]

            # Add another row with one datetime step to make the plot look better
            df.loc[df.index[-1] +
                   datetime.timedelta(minutes=env.timescale)] = df.iloc[-1]

            for port in range(cs.n_ports):
                for i, (t_arr, t_dep) in enumerate(env.port_arrival[f'{cs.id}.{port}']):
                    t_dep = t_dep + 1
                    if t_dep > len(df):
                        t_dep = len(df)
                    # x = df.index[t_arr:t_dep]
                    y = df[port].values.T[t_arr:t_dep]
                    # fill y with 0 before and after to match the length of df
                    y = np.concatenate(
                        [np.zeros(t_arr), y, np.zeros(len(df) - t_dep)])

                    plt.step(df.index,
                             y,
                             where='post',
                             color=color_list[index],
                             marker=marker_list[index],
                             label=algorithm_names[index])

            plt.title(f'Charging Station {cs.id + 1}', fontsize=24)
            plt.ylabel('SoC', fontsize=24)
            plt.ylim([0.1, 1])
            plt.xlim([env.sim_starting_date, env.sim_date])
            plt.xticks(ticks=date_range_print,
                       labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print], rotation=45,
                       fontsize=22)
            counter += 1

    plt.legend(loc='upper center', bbox_to_anchor=(1.1, -0.15),
               fancybox=True, shadow=True, ncol=5, fontsize=24)

    plt.grid(True, which='minor', axis='both')
    plt.tight_layout()

    fig_name = f'{save_path}/EV_Energy_Level.png'
    plt.savefig(fig_name, format='png',
                dpi=60, bbox_inches='tight')
    

def plot_comparable_EV_SoC_single(results_path, save_path=None, algorithm_names=None):
    '''
    This function is used to plot the SoC of the EVs in the same plot
    '''

    with open(results_path, 'rb') as f:
        replay = pickle.load(f)

    plt.close('all')
    fig, ax = plt.subplots()
    plt.rc('font', family='serif')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    
    ax.spines['left'].set_linewidth(2)
    ax.spines['bottom'].set_linewidth(2)

    plt.grid(True, which='major', axis='both')
    for index, key in enumerate(replay.keys()):        
        env = replay[key]

        date_range = pd.date_range(start=env.sim_starting_date,
                                   end=env.sim_starting_date +
                                   (env.simulation_length - 1) *
                                   datetime.timedelta(
                                       minutes=env.timescale),
                                   freq=f'{env.timescale}min')
        date_range_print = pd.date_range(start=env.sim_starting_date,
                                         end=env.sim_date,
                                         periods=7)

        color_list_map = plt.cm.get_cmap('Set1', len(replay.keys()))
        color_list = color_list_map(np.linspace(0, 1, len(replay.keys())))
        
        counter = 1
        for cs in env.charging_stations:   
            if counter != 1:
                counter += 1
                continue
            
            # plt.subplot(1, 2, counter)
            df = pd.DataFrame([], index=date_range)

            for port in range(cs.n_ports):
                df[port] = env.port_energy_level[port, cs.id, :]

            # Add another row with one datetime step to make the plot look better
            df.loc[df.index[-1] +
                   datetime.timedelta(minutes=env.timescale)] = df.iloc[-1]

            for port in range(cs.n_ports):
                for i, (t_arr, t_dep) in enumerate(env.port_arrival[f'{cs.id}.{port}']):
                    t_dep = t_dep + 1
                    if t_dep > len(df):
                        t_dep = len(df)
                    # x = df.index[t_arr:t_dep]
                    y = df[port].values.T[t_arr:t_dep]
                    # fill y with 0 before and after to match the length of df
                    y = np.concatenate(
                        [np.zeros(t_arr), y, np.zeros(len(df) - t_dep)])

                    plt.step(df.index,
                             y,
                             where='post',
                             color=color_list[index],
                             marker=marker_list[index],
                             alpha=0.8,
                             label=algorithm_names[index])

            # plt.title(f'Charging Station {cs.id + 1}', fontsize=24)
            
            if counter == 1:
                plt.ylabel('SoC', fontsize=28)
                plt.yticks(np.arange(0, 1.1, 0.2),
                           fontsize=28)
                    
            else:
                plt.yticks(fontsize=28)
                plt.yticks(np.arange(0, 1.1, 0.1),
                            labels=[' ' for d in np.arange(0, 1.1, 0.1)])            
            
            plt.ylim([0.1, 1.09])
            plt.xlim([env.sim_starting_date, env.sim_date])
            plt.xticks(ticks=date_range_print,
                       labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print],
                    #    rotation=45,
                       fontsize=28)
            counter += 1

    # plt.legend(loc='upper center', bbox_to_anchor=(0, -0.15),
    #            fancybox=True, shadow=True, ncol=, fontsize=24)

    plt.tight_layout()

    fig_name = f'{save_path}/EV_Energy_Level_single.png'
    plt.savefig(fig_name, format='png',
                dpi=60, bbox_inches='tight')

    plt.show() 
    

def plot_comparable_CS_Power(results_path, save_path=None, algorithm_names=None):
    '''
    This function is used to plot the SoC of the EVs in the same plot
    '''

    with open(results_path, 'rb') as f:
        replay = pickle.load(f)

    plt.close('all')
    fig, ax = plt.subplots()
    plt.rc('font', family='serif')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    
    ax.spines['left'].set_linewidth(2)
    ax.spines['bottom'].set_linewidth(2)

    plt.grid(True, which='major', axis='both')
    for index, key in enumerate(replay.keys()):        
        env = replay[key]

        date_range = pd.date_range(start=env.sim_starting_date,
                                   end=env.sim_starting_date +
                                   (env.simulation_length - 1) *
                                   datetime.timedelta(
                                       minutes=env.timescale),
                                   freq=f'{env.timescale}min')
        date_range_print = pd.date_range(start=env.sim_starting_date,
                                         end=env.sim_date,
                                         periods=7)

        color_list_map = plt.cm.get_cmap('Set1', len(replay.keys()))
        color_list = color_list_map(np.linspace(0, 1, len(replay.keys())))
        
        counter = 1
        for cs in env.charging_stations:   
            if counter != 1:
                counter += 1
                continue
            
            # plt.subplot(1, 2, counter)
            df = pd.DataFrame([], index=date_range)

            for port in range(cs.n_ports):
                df[port] = env.port_current[port, cs.id, :]
            
            #multiply df[port] by the voltage to get the power
            df = df * cs.voltage * math.sqrt(cs.phases) / 1000
            
            # Add another row with one datetime step to make the plot look better
            df.loc[df.index[-1] +
                   datetime.timedelta(minutes=env.timescale)] = df.iloc[-1]

            for port in range(cs.n_ports):
                for i, (t_arr, t_dep) in enumerate(env.port_arrival[f'{cs.id}.{port}']):
                    t_dep = t_dep + 1
                    if t_dep > len(df):
                        t_dep = len(df)
                    # x = df.index[t_arr:t_dep]
                    y = df[port].values.T[t_arr:t_dep]
                    # fill y with 0 before and after to match the length of df
                    y = np.concatenate(
                        [np.zeros(t_arr), y, np.zeros(len(df) - t_dep)])

                    plt.step(df.index,
                             y,
                             where='post',
                             color=color_list[index],
                             marker=marker_list[index],
                             alpha=0.8,
                             label=algorithm_names[index])

            # plt.title(f'Charging Station {cs.id + 1}', fontsize=24)
            
            if counter == 1:
                plt.ylabel('Power (kW)', fontsize=28)
                plt.yticks([-22,-11,0,11,22],
                           fontsize=28)
                    
            else:
                plt.yticks(fontsize=28)
                # plt.yticks(np.arange(0, 1.1, 0.1),
                #             labels=[' ' for d in np.arange(0, 1.1, 0.1)])            
            
            # plt.ylim([0.1, 1.09])
            plt.xlim([env.sim_starting_date, env.sim_date])
            plt.xticks(ticks=date_range_print,
                       labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print],
                    #    rotation=45,
                       fontsize=28)
            counter += 1

    # plt.legend(loc='upper center', bbox_to_anchor=(0, -0.15),
    #            fancybox=True, shadow=True, ncol=, fontsize=24)

    plt.tight_layout()

    fig_name = f'{save_path}/CS_Power_single.png'
    plt.savefig(fig_name, format='png',
                dpi=60, bbox_inches='tight')

    plt.show()   


def plot_actual_power_vs_setpoint(results_path, save_path=None, algorithm_names=None):
    
    '''
    This function is used to plot the actual power vs the setpoint power.
    It plots the behavior of each algorithm in subplots vertically.
    '''
    
    with open(results_path, 'rb') as f:
        replay = pickle.load(f)

    plt.close('all')
    plt.figure(figsize=(7, 11))
    plt.rc('font', family='serif')    

    for index, key in enumerate(replay.keys()):        
        env = replay[key]

        date_range = pd.date_range(start=env.sim_starting_date,
                                   end=env.sim_starting_date +
                                   (env.simulation_length - 1) *
                                   datetime.timedelta(
                                       minutes=env.timescale),
                                   freq=f'{env.timescale}min')
        date_range_print = pd.date_range(start=env.sim_starting_date,
                                         end=env.sim_date,
                                         periods=7)

        #plot the actual power vs the setpoint power for each algorithm in subplots                
        plt.subplot(len(replay), 1, index+1)
        plt.grid(True, which='major', axis='both')
        
        actual_power = env.current_power_usage        
        setpoints = env.power_setpoints                

        plt.step(date_range, actual_power.T, alpha=0.9, color='#00429d')
        plt.step(date_range, setpoints.T, alpha=1, color='#93003a')
        
        plt.axhline(0, color='black', lw=2)
        plt.title(f'{algorithm_names[index]}', fontsize=22)
        
        plt.yticks(fontsize=22)
        
        if index == len(replay) - 1:
            plt.xticks(ticks=date_range_print,
                       labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print],
                    #    rotation=45,
                       fontsize=22)
            # plt.xlabel('Time', fontsize=28)
        else:
            plt.xticks(ticks=date_range_print,
                       labels=[' ' for d in date_range_print])
        
        if index == len(replay) // 2:
            plt.ylabel('Power (kW)', fontsize=22)               
            
        plt.xlim([env.sim_starting_date, env.sim_date])
        plt.ylim([0, 1.1*env.current_power_usage.max()])
        
    # Put the legend under the plot in a separate axis           
    plt.legend(['Actual Power', 'Setpoint'], loc='upper center',
               bbox_to_anchor=(0.5, -0.5),
               fancybox=True, shadow=True, ncol=2, fontsize=22)
        
    plt.tight_layout()
    fig_name = f'{save_path}/Actual_vs_Setpoint_Power.png'
    plt.savefig(fig_name, format='png',
                dpi=60, bbox_inches='tight')    
    
def plot_prices(results_path, save_path=None, algorithm_names=None):
    with open(results_path, 'rb') as f:
        replay = pickle.load(f)

    plt.close('all')
    plt.figure(figsize=(7, 11))
    plt.rc('font', family='serif')
    
    keys = list(replay.keys())
    env = replay[keys[0]]
    
    date_range = pd.date_range(start=env.sim_starting_date,
                                 end=env.sim_starting_date +
                                 (env.simulation_length - 1) *
                                 datetime.timedelta(
                                      minutes=env.timescale),
                                 freq=f'{env.timescale}min')
    date_range_print = pd.date_range(start=env.sim_starting_date,
                                     end=env.sim_date,
                                     periods=7)
    
    plt.close('all')
    fig, ax = plt.subplots()
    
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['left'].set_linewidth(2)
    ax.spines['bottom'].set_linewidth(2)
    
    plt.grid(True, which='major', axis='both')    
    plt.rc('font', family='serif')
    
    charge_prices = env.charge_prices[0, :]
    discharge_prices = env.discharge_prices[0, :]
    
    plt.step(date_range, -charge_prices, alpha=0.9, color='#00429d',label='Charge Prices')
    plt.step(date_range, discharge_prices, alpha=1, color='#93003a', label='Discharge Prices')
    
    plt.xlim([env.sim_starting_date, env.sim_date])
    # plt.ylim()
    # plt.axhline(0, color='black', lw=2)    
    y_ticks = np.arange(0.150, 0.351, 0.05)
    plt.yticks(y_ticks,fontsize=28)

    plt.ylim([0.12, 0.35])
    plt.xticks(ticks=date_range_print,
               labels=[f'{d.hour:2d}:{d.minute:02d}' for d in date_range_print],
            #    rotation=45,
               fontsize=28)
    plt.ylabel('Price (€/kWh)', fontsize=28)
    
    plt.legend(fontsize=28)
    
    #show grid lines
    
    
    # plt.tight_layout()
    fig_name = f'{save_path}/Prices.png'
    plt.savefig(fig_name, format='png',
                dpi=60, bbox_inches='tight')
    plt.show()

if __name__ == "__main__":

    plot_total_power_V2G(results_path='E:/GitHub\ev2gym/results/eval_5cs_1tr_V2G_MPC_5_algos_1_exp_2024_03_03_727260/plot_results_dict.pkl',
                     save_path='E:\GitHub\ev2gym\results\eval_5cs_1tr_V2G_MPC_5_algos_1_exp_2024_03_03_727260',
                     algorithm_names=['Charge As Fast As Possible', 'OCCF V2G', 'OCCF G2V', 'eMPC V2G', 'eMPC G2V'])
    # plot_comparable_EV_SoC(results_path='plot_results_dict.pkl')
    pass



'''
This file contains the EV_Charger class, which is used to represent the EV chargers in the environment.
===================================
Author: Stavros Orfanoudakis 2023
===================================
'''

import numpy as np
import math


class EV_Charger:
    '''
    Attributes:
        - id: unique identifier of the EV charger
        - connected_bus: the bus to which the EV charger is connected
        - connected_transformer: the transformer(s) to which the EV charger is connected
        - geo_location: the geographical location of the EV charger                
        - n_ports: the number of ports of the EV charger
        - charger_type: the type of the EV charger (typ1, type2, or DC)
        - bi_directional: whether the EV charger can provide power to the grid or not
        - timescale: the timescale of the simulation (useful for determining the charging speed)
        - verbose: whether to print information about the EV charger or not

    Status variables:
        - current_power_output: the current total power output of the EV charger (positive for draining energy from the grid, negative for providing energy to the grid)
        - evs_connected: the list of EVs connected to the EV charger
        - n_ev_connected: the current number of EVs connected to the EV charger
        - current_step: the current simulation timestep

    Statistics variables:
        - total_energy_charged: the total energy charged by the EV charger  
        - total_energy_discharged: the total energy discharged by the EV charger
        - total_profits: the total profit of the EV charger
        - total_evs_served: the total number of EVs served by the EV charger
        - total_user_satisfaction: the total user satisfaction of the EV charger

    Methods:
        - step: updates the EV charger status according to the actions taken by the EVs
        - reset: resets the EV charger status to the initial state      

'''

    def __init__(self,
                 id,
                 connected_bus,
                 connected_transformer,
                 geo_location=None,
                 min_charge_current=0,  # Amperes
                 max_charge_current=56,  # Amperes
                 min_discharge_current=0,  # Amperes
                 max_discharge_current=-56,  # Amperes
                 voltage=230,  # Volts
                 n_ports=2,
                 charger_type="AC",  # AC or DC
                 phases=3,
                 timescale=5,
                 verbose=False):

        self.id = id

        # EV Charger location and grid characteristics
        self.connected_bus = connected_bus
        self.connected_transformer = connected_transformer
        self.geo_location = geo_location

        # EV Charger technical characteristics
        self.n_ports = n_ports
        self.charger_type = charger_type
        self.timescale = timescale

        self.min_charge_current = min_charge_current
        self.max_charge_current = max_charge_current
        self.min_discharge_current = min_discharge_current
        self.max_discharge_current = max_discharge_current
        self.phases = phases

        self.voltage = voltage

        # EV Charger status
        self.current_power_output = 0
        self.evs_connected = [None] * n_ports
        self.n_evs_connected = 0
        self.current_step = 0
        self.current_charge_price = 0
        self.current_discharge_price = 0
        self.current_total_amps = 0
        self.current_signal = [0]*n_ports

        # EV Charger Statistics
        self.total_energy_charged = 0
        self.total_energy_discharged = 0
        self.total_profits = 0
        self.total_evs_served = 0
        self.total_user_satisfaction = 0

        self.verbose = verbose

    def step(self, actions, charge_price, discharge_price):
        '''
        Updates the EV charger status according to the actions taken by the EVs
        Inputs:
            - actions: a list of actions taken by the EVs connected to the EV charger in the format of (current) *n_ports positive for charging negative for discharging, default is to zer if no ev is connected        
            - charge_price: the price of charging per kWh in the current timestep
            - discharge_price: the price of discharging per kWh in the current timestep

        Outputs:
            - profit: the total profit + costs of charging and discharging in the current timestep
            - user_satisfaction: a list of user satisfaction values for each EV connected to the EV charger in the current timestep
        '''
        profit = 0
        user_satisfaction = []
        self.current_power_output = 0
        self.current_total_amps = 0
        self.current_charge_price = charge_price
        self.current_discharge_price = discharge_price
        self.current_signal = []

        assert (len(actions) == self.n_ports)
        # if no EV is connected, set action to 0
        invalid_action_punishment = 0
        for i in range(len(actions)):
            if self.evs_connected[i] is None:
                actions[i] = 0
                invalid_action_punishment += 1

        # normalize actions to sum to 1 for charging surplass or -1 for discharging surplass
        if sum(actions) > 1:
            normalized_actions = [action / sum(actions) for action in actions]
        elif sum(actions) < -1:
            normalized_actions = [- action /
                                  sum(actions) for action in actions]
        else:
            normalized_actions = actions

        if self.verbose:
            print(f'CS {self.id} normalized actions: {normalized_actions}')

        # Update EVs connected to the EV charger and get profits/costs
        for i, action in enumerate(normalized_actions):
            actual_energy = 0
            action = round(action, 5)
            assert (action >= -1 and action <= 1,
                    f'Action {action} is not in range [-1,1]')

            amps = 0
            if action == 0 and self.evs_connected[i] is not None:

                actual_energy, actual_amps = self.evs_connected[i].step(
                    amps, self.voltage)

            elif action > 0:
                amps = action * self.max_charge_current
                if amps < self.min_charge_current-0.01:
                    amps = 0

                actual_energy, actual_amps = self.evs_connected[i].step(
                    amps,
                    self.voltage,
                    phases=self.phases,
                    type=self.charger_type)

                profit += abs(actual_energy) * charge_price
                self.total_energy_charged += abs(actual_energy)
                self.current_power_output += actual_energy * 60/self.timescale
                self.current_total_amps += actual_amps

            elif action < 0:
                amps = action * abs(self.max_discharge_current)
                if amps > self.min_discharge_current-0.01:
                    amps = self.min_discharge_current

                actual_energy, actual_amps = self.evs_connected[i].step(
                    amps,
                    self.voltage,
                    phases=self.phases,
                    type=self.charger_type)

                profit += abs(actual_energy) * discharge_price
                self.total_energy_discharged += abs(actual_energy)
                self.current_power_output += actual_energy * 60/self.timescale
                self.current_total_amps += actual_amps

            # print(f'CS {self.id} port {i} action {action} amps {amps} energy {actual_energy} total_amps {self.current_total_amps}')
            self.current_signal.append(amps)

            if self.current_total_amps - 0.0001 > self.max_charge_current:
                raise Exception(
                    f'sum of amps {self.current_total_amps} is higher than max charge current {self.max_charge_current}')

        self.total_profits += profit

        departing_evs = []
        # Check if EVs are departing
        for i, ev in enumerate(self.evs_connected):
            if ev is not None:
                if ev.is_departing(self.current_step) is not None:
                    # calculate battery degradation
                    # _,_ = ev.get_battery_degradation()
                    self.evs_connected[i] = None
                    self.n_evs_connected -= 1
                    self.total_evs_served += 1
                    ev_user_satisfaction = ev.get_user_satisfaction()
                    self.total_user_satisfaction += ev_user_satisfaction
                    user_satisfaction.append(ev_user_satisfaction)
                    departing_evs.append(ev)
                    if self.verbose:
                        print(f'- EV {ev.id} is departing from CS {self.id}' +
                              f' port {i}'
                              f' with user satisfaction {ev_user_satisfaction}' +
                              f' (SoC: {ev.get_soc()*100: 6.1f}%)')

        self.current_step += 1

        return profit, user_satisfaction, invalid_action_punishment, departing_evs

    def __str__(self) -> str:

        if self.total_evs_served == 0:
            user_satisfaction_str = ' Avg. Sat.:  - '
        else:
            user_satisfaction_str = f' Avg. Sat.: {self.get_avg_user_satisfaction()*100: 3.1f}%'

        # f' ({self.current_step*self.timescale: 3d} mins)' + \
        return f'CS{self.id:3d}: ' + \
            f' Served {self.total_evs_served:4d} EVs' + \
            user_satisfaction_str + \
            f' in {self.current_step: 4d} steps' + \
            f' |{self.total_profits: 7.1f} € |' + \
            f' +{self.total_energy_charged: 5.1f} /' + \
            f' -{self.total_energy_discharged: 5.1f} kWh'

    def get_max_power(self):
        return self.max_charge_current * self.voltage * math.sqrt(self.phases) / 1000

    def get_min_charge_power(self):
        return self.min_charge_current * self.voltage * math.sqrt(self.phases) / 1000

    def get_min_power(self):
        return self.max_discharge_current * self.voltage * math.sqrt(self.phases) / 1000

    def get_avg_user_satisfaction(self):
        if self.total_evs_served == 0:
            return 0
        else:
            return self.total_user_satisfaction / self.total_evs_served

    def spawn_ev(self, ev):
        '''Adds an EV to the list of EVs connected to the EV charger
        Inputs:
            - ev: the EV to be added to the list of EVs connected to the EV charger
        '''
        assert (self.n_evs_connected < self.n_ports)

        index = self.evs_connected.index(None)
        ev.id = index
        self.evs_connected[index] = ev
        self.n_evs_connected += 1

        if self.verbose:
            print(f'+ EV connected to Charger {self.id} at port {index}' +
                  f' leaving at {ev.time_of_departure}' +
                  f' SoC {ev.get_soc()*100:.1f}%')

        return index

    def reset(self):
        '''Resets the EV charger status to the initial state'''

        # EV Charger status
        self.current_power_output = 0
        self.current_total_amps = 0
        self.evs_connected = [None] * self.n_ports
        self.n_evs_connected = 0
        self.current_step = 0

        # EV Charger Statistics
        self.total_energy_charged = 0
        self.total_energy_discharged = 0
        self.total_profits = 0
        self.total_evs_served = 0
        self.total_user_satisfaction = 0
